import unittest
from functools import partial

from curtsies.configfile_keynames import keymap
from curtsies.events import CURTSIES_NAMES


class TestKeymap(unittest.TestCase):
    def config(self, mapping, curtsies):
        curtsies_names = keymap[mapping]
        self.assertTrue(
            curtsies in CURTSIES_NAMES.values(), "%r is not a curtsies name" % curtsies
        )
        self.assertTrue(
            curtsies in curtsies_names,
            "config name %r does not contain %r, just %r"
            % (mapping, curtsies, curtsies_names),
        )

    def test_simple(self):
        self.config("M-m", "<Esc+m>")
        self.config("M-m", "<Esc+m>")
        self.config("C-m", "<Ctrl-m>")
        self.config("C-[", "<ESC>")
        self.config("C-\\", "<Ctrl-\\>")
        self.config("C-]", "<Ctrl-]>")
        self.config("C-^", "<Ctrl-6>")
        self.config("C-_", "<Ctrl-/>")  # ??? for bpython compatibility
        self.config("F1", "<F1>")



import unittest
from functools import partial

from curtsies import events


class TestChrByte(unittest.TestCase):
    def test_simple(self):
        """0-255 can be turned into bytes"""
        for i in range(256):
            s = eval('b"\\x%s"' % ("0" + hex(i)[2:])[-2:])
            self.assertEqual(s, events.chr_byte(i))

    def test_helpers(self):
        self.assertEqual(events.chr_byte(97), b"a")
        self.assertEqual(events.chr_uni(97), "a")


class TestCurtsiesNames(unittest.TestCase):
    def spot_check(self):
        self.assertEqual(events.CURTSIES_NAMES[b"\x1b\x08"], "<Esc+Ctrl-H>")
        self.assertEqual(events.CURTSIES_NAMES[b"\x00"], "<Ctrl-SPACE>")
        self.assertEqual(events.CURTSIES_NAMES[b"\xea"], "<Meta-j>")

    def test_all_values_unicode(self):
        for seq, e in events.CURTSIES_NAMES.items():
            self.assertEqual(type(seq), bytes)

    def test_all_keys_bytes(self):
        for seq, e in events.CURTSIES_NAMES.items():
            self.assertEqual(type(e), str)


class TestDecodable(unittest.TestCase):
    def test_simple(self):
        self.assertTrue(events.decodable(b"d", "utf-8"))
        self.assertFalse(events.decodable(b"\xfe", "utf-8"))  # 254 is off limits


class TestGetKey(unittest.TestCase):
    def test_utf8_full(self):
        get_utf_full = partial(
            events.get_key, encoding="utf-8", keynames="curtsies", full=True
        )
        self.assertEqual(get_utf_full([b"h"]), "h")
        self.assertEqual(get_utf_full([b"\x1b", b"["]), "<Esc+[>")
        self.assertRaises(UnicodeDecodeError, get_utf_full, [b"\xfe\xfe"])
        self.assertRaises(ValueError, get_utf_full, "a")

    def test_utf8(self):
        get_utf = partial(
            events.get_key, encoding="utf-8", keynames="curtsies", full=False
        )
        self.assertEqual(get_utf([b"h"]), "h")
        self.assertEqual(get_utf([b"\x1b", b"["]), None)
        self.assertEqual(get_utf([b"\xe2"]), None)
        self.assertRaises(ValueError, get_utf, "a")

    def test_multibyte_utf8(self):
        get_utf = partial(
            events.get_key, encoding="utf-8", keynames="curtsies", full=False
        )
        self.assertEqual(get_utf([b"\xc3"]), None)
        self.assertEqual(get_utf([b"\xe2"]), None)
        self.assertEqual(get_utf([b"\xe2"], full=True), "<Meta-b>")
        self.assertEqual(get_utf([b"\xc3", b"\x9f"]), "ß")
        self.assertEqual(get_utf([b"\xe2"]), None)
        self.assertEqual(get_utf([b"\xe2", b"\x88"]), None)
        self.assertEqual(get_utf([b"\xe2", b"\x88", b"\x82"]), "∂")

    def test_sequences_without_names(self):
        get_utf = partial(
            events.get_key, encoding="utf-8", keynames="curtsies", full=False
        )
        self.assertEqual(get_utf([b"\xc3"], full=True), "<Meta-C>")
        self.assertEqual(get_utf([b"\xc3"], full=True, keynames="curses"), "xC3")

    def test_key_names(self):
        "Every key sequence with a Curses name should have a Curtsies name too."
        self.assertTrue(
            set(events.CURTSIES_NAMES).issuperset(set(events.CURSES_NAMES)),
            set(events.CURSES_NAMES) - set(events.CURTSIES_NAMES),
        )


class TestGetKeyAscii(unittest.TestCase):
    def test_full(self):
        get_ascii_full = partial(
            events.get_key, encoding="ascii", keynames="curtsies", full=True
        )
        self.assertEqual(get_ascii_full([b"a"]), "a")
        self.assertEqual(get_ascii_full([b"\xe1"]), "<Meta-a>")
        self.assertEqual(get_ascii_full([b"\xe1"], keynames="curses"), "xE1")

    def test_simple(self):
        get_ascii_full = partial(events.get_key, encoding="ascii", keynames="curtsies")
        self.assertEqual(get_ascii_full([b"a"]), "a")
        self.assertEqual(get_ascii_full([b"\xe1"]), "<Meta-a>")
        self.assertEqual(get_ascii_full([b"\xe1"], keynames="curses"), "xE1")


class TestUnknownEncoding(unittest.TestCase):
    def test_simple(self):
        get_utf16 = partial(events.get_key, encoding="utf16", keynames="curtsies")
        self.assertEqual(get_utf16([b"a"]), None)
        self.assertEqual(get_utf16([b"a"], full=True), None)
        self.assertEqual(get_utf16([b"\xe1"]), None)
        self.assertEqual(get_utf16([b"\xe1"], full=True), "<Meta-a>")


class TestSpecialKeys(unittest.TestCase):
    def test_simple(self):
        seq = [b"\x1b", b"[", b"1", b";", b"9", b"C"]
        self.assertEqual(
            [events.get_key(seq[:i], encoding="utf8") for i in range(1, len(seq) + 1)],
            [None, None, None, None, None, "<Esc+RIGHT>"],
        )


class TestPPEvent(unittest.TestCase):
    def test(self):
        self.assertEqual(events.pp_event("a"), "a")



"""Events for keystrokes and other input events"""
import codecs
import encodings
import itertools
import sys
from typing import Optional, List, Union

from .termhelpers import Termmode
from .curtsieskeys import CURTSIES_NAMES as special_curtsies_names

chr_byte = lambda i: chr(i).encode("latin-1")
chr_uni = chr


CURTSIES_NAMES = {}
control_chars = {chr_byte(i): "<Ctrl-%s>" % chr(i + 0x60) for i in range(0x00, 0x1B)}
CURTSIES_NAMES.update(control_chars)
for i in range(0x00, 0x80):
    CURTSIES_NAMES[b"\x1b" + chr_byte(i)] = "<Esc+%s>" % chr(i)
for i in range(0x00, 0x1B):  # Overwrite the control keys with better labels
    CURTSIES_NAMES[b"\x1b" + chr_byte(i)] = "<Esc+Ctrl-%s>" % chr(i + 0x40)
for i in range(0x00, 0x80):
    CURTSIES_NAMES[chr_byte(i + 0x80)] = "<Meta-%s>" % chr(i)
for i in range(0x00, 0x1B):  # Overwrite the control keys with better labels
    CURTSIES_NAMES[chr_byte(i + 0x80)] = "<Meta-Ctrl-%s>" % chr(i + 0x40)

CURTSIES_NAMES.update(special_curtsies_names)

CURSES_NAMES = {
    b"\x1bOP": "KEY_F(1)",
    b"\x1bOQ": "KEY_F(2)",
    b"\x1bOR": "KEY_F(3)",
    b"\x1bOS": "KEY_F(4)",
    b"\x1b[15~": "KEY_F(5)",
    b"\x1b[17~": "KEY_F(6)",
    b"\x1b[18~": "KEY_F(7)",
    b"\x1b[19~": "KEY_F(8)",
    b"\x1b[20~": "KEY_F(9)",
    b"\x1b[21~": "KEY_F(10)",
    b"\x1b[23~": "KEY_F(11)",
    b"\x1b[24~": "KEY_F(12)",
    # see bpython #626
    b"\x1b[11~": "KEY_F(1)",
    b"\x1b[12~": "KEY_F(2)",
    b"\x1b[13~": "KEY_F(3)",
    b"\x1b[14~": "KEY_F(4)",
    b"\x1b[A": "KEY_UP",
    b"\x1b[B": "KEY_DOWN",
    b"\x1b[C": "KEY_RIGHT",
    b"\x1b[D": "KEY_LEFT",
    b"\x1b[F": "KEY_END",  # https://github.com/bpython/bpython/issues/490
    b"\x1b[H": "KEY_HOME",  # https://github.com/bpython/bpython/issues/490
    b"\x08": "KEY_BACKSPACE",
    b"\x1b[Z": "KEY_BTAB",
    # see curtsies #78 - taken from https://github.com/jquast/blessed/blob/e9ad7b85dfcbbba49010ab8c13e3a5920d81b010/blessed/keyboard.py#L409
    b"\x1b[1~": "KEY_FIND",  # find
    b"\x1b[2~": "KEY_IC",  # insert (0)
    b"\x1b[3~": "KEY_DC",  # delete (.), "Execute"
    b"\x1b[4~": "KEY_SELECT",  # select
    b"\x1b[5~": "KEY_PPAGE",  # pgup   (9)
    b"\x1b[6~": "KEY_NPAGE",  # pgdown (3)
    b"\x1b[7~": "KEY_HOME",  # home
    b"\x1b[8~": "KEY_END",  # end
    b"\x1b[OA": "KEY_UP",  # up     (8)
    b"\x1b[OB": "KEY_DOWN",  # down   (2)
    b"\x1b[OC": "KEY_RIGHT",  # right  (6)
    b"\x1b[OD": "KEY_LEFT",  # left   (4)
    b"\x1b[OF": "KEY_END",  # end    (1)
    b"\x1b[OH": "KEY_HOME",  # home   (7)
}

KEYMAP_PREFIXES = set()
for table in (CURSES_NAMES, CURTSIES_NAMES):
    for k in table:
        if k.startswith(b"\x1b"):
            for i in range(1, len(k)):
                KEYMAP_PREFIXES.add(k[:i])

MAX_KEYPRESS_SIZE = max(
    len(seq) for seq in itertools.chain(CURSES_NAMES.keys(), CURTSIES_NAMES.keys())
)


class Event:
    pass


class ScheduledEvent(Event):
    """Event scheduled for a future time.

    args:
        when (float): unix time in seconds for which this event is scheduled

    Custom events that occur at a specific time in the future should
    be subclassed from ScheduledEvent."""

    def __init__(self, when: float) -> None:
        self.when = when


class WindowChangeEvent(Event):
    def __init__(
        self, rows: int, columns: int, cursor_dy: Optional[int] = None
    ) -> None:
        self.rows = rows
        self.columns = columns
        self.cursor_dy = cursor_dy

    x = width = property(lambda self: self.columns)
    y = height = property(lambda self: self.rows)

    def __repr__(self) -> str:
        return "<WindowChangeEvent (%d, %d)%s>" % (
            self.rows,
            self.columns,
            "" if self.cursor_dy is None else " cursor_dy: %d" % self.cursor_dy,
        )

    @property
    def name(self) -> str:
        return "<WindowChangeEvent>"


class SigIntEvent(Event):
    """Event signifying a SIGINT"""

    def __repr__(self) -> str:
        return "<SigInt Event>"

    @property
    def name(self) -> str:
        return repr(self)


class PasteEvent(Event):
    """Multiple keypress events combined, likely from copy/paste.

    The events attribute contains a list of keypress event strings.
    """

    def __init__(self) -> None:
        self.events: List[str] = []

    def __repr__(self) -> str:
        return "<Paste Event with data: %r>" % self.events

    @property
    def name(self) -> str:
        return repr(self)


def decodable(seq: bytes, encoding: str) -> bool:
    try:
        u = seq.decode(encoding)
    except UnicodeDecodeError:
        return False
    else:
        return True


def get_key(
    bytes_: List[bytes], encoding: str, keynames: str = "curtsies", full: bool = False
) -> Optional[str]:
    """Return key pressed from bytes_ or None

    Return a key name or None meaning it's an incomplete sequence of bytes
    (more bytes needed to determine the key pressed)

    encoding is how the bytes should be translated to unicode - it should
    match the terminal encoding.

    keynames is a string describing how keys should be named:

    * curtsies uses unicode strings like <F8>

    * curses uses unicode strings similar to those returned by
      the Python ncurses window.getkey function, like KEY_F(8),
      plus a nonstandard representation of meta keys (bytes 128-255)
      because returning the corresponding unicode code point would be
      indistinguishable from the multibyte sequence that encodes that
      character in the current encoding

    * bytes returns the original bytes from stdin (NOT unicode)

    if full, match a key even if it could be a prefix to another key
    (useful for detecting a plain escape key for instance, since
    escape is also a prefix to a bunch of char sequences for other keys)

    Events are subclasses of Event, or unicode strings

    Precondition: get_key(prefix, keynames) is None for all proper prefixes of
    bytes. This means get_key should be called on progressively larger inputs
    (for 'asdf', first on 'a', then on 'as', then on 'asd' - until a non-None
    value is returned)
    """
    if not all(isinstance(c, type(b"")) for c in bytes_):
        raise ValueError("get key expects bytes, got %r" % bytes_)  # expects raw bytes
    if keynames not in ["curtsies", "curses", "bytes"]:
        raise ValueError("keynames must be one of 'curtsies', 'curses' or 'bytes'")
    seq = b"".join(bytes_)
    if len(seq) > MAX_KEYPRESS_SIZE:
        raise ValueError("unable to decode bytes %r" % seq)

    def key_name() -> str:
        if keynames == "curses":
            # may not be here (and still not decodable) curses names incomplete
            if seq in CURSES_NAMES:
                return CURSES_NAMES[seq]

            # Otherwise, there's no special curses name for this
            try:
                # for normal decodable text or a special curtsies sequence with bytes that can be decoded
                return seq.decode(encoding)
            except UnicodeDecodeError:
                # this sequence can't be decoded with this encoding, so we need to represent the bytes
                if len(seq) == 1:
                    return "x%02X" % ord(seq)
                    # TODO figure out a better thing to return here
                else:
                    raise NotImplementedError(
                        "are multibyte unnameable sequences possible?"
                    )
                    return "bytes: " + "-".join(
                        "x%02X" % ord(seq[i : i + 1]) for i in range(len(seq))
                    )
                    # TODO if this isn't possible, return multiple meta keys as a paste event if paste events enabled
        elif keynames == "curtsies":
            if seq in CURTSIES_NAMES:
                return CURTSIES_NAMES[seq]
            return seq.decode(
                encoding
            )  # assumes that curtsies names are a subset of curses ones
        else:
            assert keynames == "bytes"
            return seq  # type: ignore

    key_known = seq in CURTSIES_NAMES or seq in CURSES_NAMES or decodable(seq, encoding)

    if full and key_known:
        return key_name()
    elif seq in KEYMAP_PREFIXES or could_be_unfinished_char(seq, encoding):
        return None  # need more input to make up a full keypress
    elif key_known:
        return key_name()
    else:
        # this will raise a unicode error (they're annoying to raise ourselves)
        seq.decode(encoding)
        assert False, "should have raised an unicode decode error"


def could_be_unfinished_char(seq: bytes, encoding: str) -> bool:
    """Whether seq bytes might create a char in encoding if more bytes were added"""
    if decodable(seq, encoding):
        return False  # any sensible encoding surely doesn't require lookahead (right?)
        # (if seq bytes encoding a character, adding another byte shouldn't also encode something)

    if codecs.getdecoder("utf8") is codecs.getdecoder(encoding):
        return could_be_unfinished_utf8(seq)
    elif codecs.getdecoder("ascii") is codecs.getdecoder(encoding):
        return False
    else:
        return True  # We don't know, it could be


def could_be_unfinished_utf8(seq: bytes) -> bool:
    # http://en.wikipedia.org/wiki/UTF-8#Description

    # fmt: off
    if   ord(seq[0:1]) & 0b10000000 == 0b10000000 and len(seq) < 1: return True
    elif ord(seq[0:1]) & 0b11100000 == 0b11000000 and len(seq) < 2: return True
    elif ord(seq[0:1]) & 0b11110000 == 0b11100000 and len(seq) < 3: return True
    elif ord(seq[0:1]) & 0b11111000 == 0b11110000 and len(seq) < 4: return True
    elif ord(seq[0:1]) & 0b11111100 == 0b11111000 and len(seq) < 5: return True
    elif ord(seq[0:1]) & 0b11111110 == 0b11111100 and len(seq) < 6: return True
    else: return False
    # fmt: on


def pp_event(seq: Union[Event, str]) -> Union[str, bytes]:
    """Returns pretty representation of an Event or keypress"""

    if isinstance(seq, Event):
        return str(seq)

    # Get the original sequence back if seq is a pretty name already
    rev_curses = {v: k for k, v in CURSES_NAMES.items()}
    rev_curtsies = {v: k for k, v in CURTSIES_NAMES.items()}
    bytes_seq: Optional[bytes] = None
    if seq in rev_curses:
        bytes_seq = rev_curses[seq]
    elif seq in rev_curtsies:
        bytes_seq = rev_curtsies[seq]

    if bytes_seq:
        pretty = curtsies_name(bytes_seq)
        if pretty != seq:
            return pretty
    return repr(seq).lstrip("u")[1:-1]


def curtsies_name(seq: bytes) -> Union[str, bytes]:
    return CURTSIES_NAMES.get(seq, seq)


def try_keys() -> None:
    print(
        "press a bunch of keys (not at the same time, but you can hit them pretty quickly)"
    )
    import tty
    import termios
    import fcntl
    import os
    from .termhelpers import Cbreak

    def ask_what_they_pressed(seq: bytes, Normal: Termmode) -> None:
        print("Unidentified character sequence!")
        with Normal:
            while True:
                r = input("type 'ok' to prove you're not pounding keys ")
                if r.lower().strip() == "ok":
                    break
        while True:
            print(f"Press the key that produced {seq!r} again please")
            retry = os.read(sys.stdin.fileno(), 1000)
            if seq == retry:
                break
            print("nope, that wasn't it")
        with Normal:
            name = input("Describe in English what key you pressed: ")
            f = open("keylog.txt", "a")
            f.write(f"{seq!r} is called {name}\n")
            f.close()
            print(
                "Thanks! Please open an issue at https://github.com/bpython/curtsies/issues"
            )
            print(
                "or email thomasballinger@gmail.com. Include this terminal history or keylog.txt."
            )
            print("You can keep pressing keys")

    with Cbreak(sys.stdin) as NoCbreak:
        while True:
            try:
                chars = os.read(sys.stdin.fileno(), 1000)
                print("---")
                print(repr(chars))
                if chars in CURTSIES_NAMES:
                    print(CURTSIES_NAMES[chars])
                elif len(chars) == 1:
                    print("literal")
                else:
                    print("unknown!!!")
                    ask_what_they_pressed(chars, NoCbreak)
            except OSError:
                pass


if __name__ == "__main__":
    try_keys()



from curtsies.input import *

def main():
    """
    Reads and returns user input; after 2 seconds, 1 second, .5 second 
    and .2 second, respectively, an event -- user input -- is printed, 
    or "None" is printed if no user input is received.
    """
    with Input() as input_generator:
        print(repr(input_generator.send(2)))
        print(repr(input_generator.send(1)))
        print(repr(input_generator.send(.5)))
        print(repr(input_generator.send(.2)))
        for e in input_generator:
            print(repr(e))
            if e == '<ESC>':
                break
if __name__ == '__main__':
    main()



import sys
import signal

from curtsies import input, Cbreak, FullscreenWindow, fmtstr

def fullscreen_winch_with_input():
    """
    Monitors user input as well as screen size and acknowledges changes to both.
    """
    print('this should be just off-screen')
    w = FullscreenWindow(sys.stdout)
    def sigwinch_handler(signum, frame):
        print('sigwinch! Changed from {!r} to {!r}'.format((rows, columns), (w.height, w.width)))
    signal.signal(signal.SIGWINCH, sigwinch_handler)
    with w:
        with Cbreak(sys.stdin):
            for e in input.Input():
                rows, columns = w.height, w.width
                a = [fmtstr(((f'.{rows}x{columns}.{e!r}.') * rows)[:columns]) for row in range(rows)]
                w.render_to_terminal(a)

                if e == '<ESC>':
                    break

if __name__ == '__main__':
    fullscreen_winch_with_input()

